package com.processmininguc.feedbackrepair.models;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XLog;
import org.jbpt.algo.tree.rpst.RPST;
import org.jbpt.algo.tree.rpst.RPSTNode;
import org.jbpt.graph.DirectedEdge;
import org.jbpt.hypergraph.abs.Vertex;
import org.jbpt.utils.IOUtils;
import org.jgraph.graph.AttributeMap.SerializablePoint2D;
import org.jgraph.graph.GraphConstants;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.models.heuristics.HeuristicsNetGraph;
import org.processmining.models.heuristics.impl.HNSet;
import org.processmining.models.heuristics.impl.HNSubSet;
import org.processmining.models.jgraph.ProMGraphModel;
import org.processmining.models.jgraph.elements.ProMGraphPort;

import com.jgraph.layout.JGraphFacade;
import com.jgraph.layout.JGraphLayoutProgress;
import com.jgraph.layout.hierarchical.JGraphHierarchicalLayout;
import com.processmininguc.feedbackrepair.parameters.RepairHeuristicNetBasedOnFeedbackConfiguration;

public class FeedbackRepair {
	
	private HeuristicsNet net;
	private HeuristicsNetGraph graph;
	private JComponent jpanel;
	private PluginContext context;
	private String[] hashToNode;
	private static XLog log;

	public FeedbackRepair(HeuristicsNet net, XLog log, RepairHeuristicNetBasedOnFeedbackConfiguration configuration) {
		// TODO Auto-generated constructor stub
		this.net = net;
		this.graph = new HeuristicsNetGraph(net, "Heuristics Net", false);
		FeedbackRepair.log = log;
	}
	
	public HeuristicsNet getNet() {
		return this.net;
	}
	
	public HeuristicsNetGraph getGraph() {
		return this.graph;
	}
	
	public JComponent getGraphPanel(PluginContext context) {
		this.context = context;
		JComponent graph = visualizeGraph(this.graph, this.net, context.getProgress());
		this.jpanel = graph;
		return graph;
	}
	
	public static HeuristicsNet ModifyNet(HeuristicsNet net, FeedbackRepairOperation op, int src, int dst) {
		long start = System.nanoTime();
		HeuristicsNet n;
		switch(op) {
			case FOLLOWS:
				n = makeFollowing(net, src, dst);
				break;
			case PARALLEL:
				n = makeParallel(net, src, dst);
				break;
			case INDIFFERENT:
				n = eraseEdge(net, src, dst);
				//n = makeIndifferent(net, src, dst);
				break;
			default:
				n = net;
				break;
		}
		long end = System.nanoTime();
		logger(end - start, op.toString(), "NETMOD");
		return n;
	}
	
	public static HeuristicsNet makeParallel(HeuristicsNet net, int src, int dst) {
		
		HNSet src_inputset = net.getInputSet(src);
		HNSet dst_inputset = net.getInputSet(dst);
		HNSet custom_inputset = new HNSet();
		
		HNSet src_outputset = net.getOutputSet(src);
		HNSet dst_outputset = net.getOutputSet(dst);
		HNSet custom_outputset = new HNSet();
		
		for(int i = 0; i < src_inputset.size(); i++) {
			HNSubSet ss = src_inputset.get(i);
			if(ss.contains(src)) {
				ss.remove(src);
			}
			if(ss.contains(dst)) {
				ss.remove(dst);
			}
			custom_inputset.add(ss);
		}
		for(int i = 0; i < dst_inputset.size(); i++) {
			HNSubSet ss = dst_inputset.get(i);
			if(ss.contains(src)) {
				ss.remove(src);
			}
			if(ss.contains(dst)) {
				ss.remove(dst);
			}
			custom_inputset.add(ss);
		}
		for(int i = 0; i < src_outputset.size(); i++) {
			HNSubSet ss = src_outputset.get(i);
			if(ss.contains(src)) {
				ss.remove(src);
			}
			if(ss.contains(dst)) {
				ss.remove(dst);
			}
			custom_outputset.add(ss);
		}
		for(int i = 0; i < dst_outputset.size(); i++) {
			HNSubSet ss = dst_outputset.get(i);
			if(ss.contains(src)) {
				ss.remove(src);
			}
			if(ss.contains(dst)) {
				ss.remove(dst);
			}
			custom_outputset.add(ss);
		}
		
		net.setInputSet(src, custom_inputset);
		net.setInputSet(dst, custom_inputset);
		net.setOutputSet(src, custom_outputset);
		net.setOutputSet(dst, custom_outputset);
		
		int size = net.getActivitiesMappingStructures().getXEventClasses().size();
		
		//Loop for each activity on the net
		for(int i = 0; i < size; i++) {
			HNSet inputset = net.getInputSet(i);
			HNSet outputset = net.getOutputSet(i);
			
			HNSet new_inputset = new HNSet();
			HNSet new_outputset = new HNSet();
			
			int inputset_size = inputset.size();
			int outputset_size = outputset.size();
			
//			//Special case where the current activity is the source activity
//			if(i == src) {
//				//Loop for each set on the activity input set
//				for(int j = 0; j < inputset_size; j++) {
//					if(inputset.get(j).contains(dst)) {
//						HNSubSet ss = inputset.get(j);
//						ss.remove(dst);
//						new_inputset.add(ss);
//					}
//					else {
//						HNSubSet ss = inputset.get(j);
//						new_inputset.add(ss);
//					}
//				}
//				//Loop for each set on the activity output set
//				for(int k = 0; k < outputset_size; k++) {
//					if(outputset.get(k).contains(dst)) {
//						HNSubSet ss = outputset.get(k);
//						ss.remove(dst);
//						new_outputset.add(ss);
//					}
//					else {
//						HNSubSet ss = outputset.get(k);
//						new_outputset.add(ss);
//					}
//				}
//			}
//			//Special case where the current activity is the destination activity
//			if(i == dst) {
//				//Loop for each set on the activity input set
//				for(int j = 0; j < inputset_size; j++) {
//					if(inputset.get(j).contains(src)) {
//						HNSubSet ss = inputset.get(j);
//						ss.remove(src);
//						new_inputset.add(ss);
//					}
//					else {
//						HNSubSet ss = inputset.get(j);
//						new_inputset.add(ss);
//					}
//				}
//				//Loop for each set on the activity output set
//				for(int k = 0; k < outputset_size; k++) {
//					if(outputset.get(k).contains(src)) {
//						HNSubSet ss = outputset.get(k);
//						ss.remove(src);
//						new_outputset.add(ss);
//					}
//					else {
//						HNSubSet ss = outputset.get(k);
//						new_outputset.add(ss);
//					}
//				}
//			}
			//General scenario
			if(i != src || i != dst) {
				//Loop for each set on the activity input set
				for(int j = 0; j < inputset_size; j++) {
					if(inputset.get(j).contains(src) && !inputset.get(j).contains(dst)) {
						HNSubSet ss = inputset.get(j);
						ss.add(dst);
						new_inputset.add(ss);
					}
					else if(inputset.get(j).contains(dst) && !inputset.get(j).contains(src)) {
						HNSubSet ss = inputset.get(j);
						ss.add(src);
						new_inputset.add(ss);
					}
					else {
						HNSubSet ss = inputset.get(j);
						new_inputset.add(ss);
					}
				}
				//Loop for each set on the activity output set
				for(int k = 0; k < outputset_size; k++) {
					if(outputset.get(k).contains(src) && !outputset.get(k).contains(dst)) {
						HNSubSet ss = outputset.get(k);
						ss.add(dst);
						new_outputset.add(ss);
					}
					else if(outputset.get(k).contains(dst) && !outputset.get(k).contains(src)) {
						HNSubSet ss = outputset.get(k);
						ss.add(src);
						new_outputset.add(ss);
					}
					else {
						HNSubSet ss = outputset.get(k);
						new_outputset.add(ss);
					}
				}
			}
			net.setInputSet(i, new_inputset);
			net.setOutputSet(i, new_outputset);
		}
		
		return net;
	}
	
	public static HeuristicsNet makeFollowing(HeuristicsNet net, int src, int dst) {
		
		HNSet src_outputset = net.getOutputSet(src);
		HNSet dst_inputset = net.getInputSet(dst);
		
		boolean contains_dst = false;
		for(int i = 0; i < src_outputset.size(); i++) {
			HNSubSet ss = src_outputset.get(i);
			if(ss.contains(dst)) {
				contains_dst = true;
				break;
			}
		}
		if(!contains_dst) {
			HNSubSet ss = new HNSubSet();
			ss.add(dst);
			src_outputset.add(ss);
		}
		
		boolean contains_src = false;
		for(int i = 0; i < dst_inputset.size(); i++) {
			HNSubSet ss = dst_inputset.get(i);
			if(ss.contains(src)) {
				contains_src = true;
				break;
			}
		}
		if(!contains_src) {
			HNSubSet ss = new HNSubSet();
			ss.add(src);
			dst_inputset.add(ss);
		}
		
		net.setOutputSet(src, src_outputset);
		net.setInputSet(dst, dst_inputset);
		
		int size = net.getActivitiesMappingStructures().getXEventClasses().size();
		
		//Loop for each activity on the net
		for(int i = 0; i < size; i++) {
			HNSet inputset = net.getInputSet(i);
			HNSet outputset = net.getOutputSet(i);
			
			HNSet new_inputset = new HNSet();
			HNSet new_outputset = new HNSet();
			
			int inputset_size = inputset.size();
			int outputset_size = outputset.size();

			//General scenario
			if(i != src || i != dst) {
				//Loop for each set on the activity input set
				for(int j = 0; j < inputset_size; j++) {
					if(inputset.get(j).contains(dst) && inputset.get(j).contains(src)) {
						HNSubSet ss = inputset.get(j);
						ss.remove(src);
						new_inputset.add(ss);
					}
					else {
						HNSubSet ss = inputset.get(j);
						new_inputset.add(ss);
					}
				}
				//Loop for each set on the activity output set
				for(int k = 0; k < outputset_size; k++) {
					if(outputset.get(k).contains(src) && outputset.get(k).contains(dst)) {
						HNSubSet ss = outputset.get(k);
						ss.remove(dst);
						new_outputset.add(ss);
					}
					else {
						HNSubSet ss = outputset.get(k);
						new_outputset.add(ss);
					}
				}
			}
			net.setOutputSet(i, new_outputset);
			net.setInputSet(i, new_inputset);
		}
		
		return net;
	}
	
	public static HeuristicsNet makeIndifferent(HeuristicsNet net, int src, int dst) {
		
		HNSet src_outputset = net.getOutputSet(src);
		HNSet dst_inputset = net.getInputSet(dst);
		
		boolean contains_dst = false;
		for(int i = 0; i < src_outputset.size(); i++) {
			HNSubSet ss = src_outputset.get(i);
			if(ss.contains(dst)) {
				contains_dst = true;
				break;
			}
		}
		if(!contains_dst) {
			HNSubSet ss = new HNSubSet();
			ss.add(dst);
			src_outputset.add(ss);
		}
		
		boolean contains_src = false;
		for(int i = 0; i < dst_inputset.size(); i++) {
			HNSubSet ss = dst_inputset.get(i);
			if(ss.contains(src)) {
				contains_src = true;
				break;
			}
		}
		if(!contains_src) {
			HNSubSet ss = new HNSubSet();
			ss.add(src);
			dst_inputset.add(ss);
		}
		
		net.setOutputSet(src, src_outputset);
		net.setInputSet(dst, dst_inputset);
		
		int size = net.getActivitiesMappingStructures().getXEventClasses().size();
		
		//Loop for each activity on the net
		for(int i = 0; i < size; i++) {
			HNSet inputset = net.getInputSet(i);
			HNSet outputset = net.getOutputSet(i);
			
			HNSet new_inputset = new HNSet();
			HNSet new_outputset = new HNSet();
			
			int inputset_size = inputset.size();
			int outputset_size = outputset.size();

			//General scenario
			if(i != src || i != dst) {
				//Loop for each set on the activity input set
				for(int j = 0; j < inputset_size; j++) {
					if(inputset.get(j).contains(dst) && inputset.get(j).contains(src)) {
						HNSubSet ss = inputset.get(j);
						ss.remove(src);
						new_inputset.add(ss);
					}
					else {
						HNSubSet ss = inputset.get(j);
						new_inputset.add(ss);
					}
				}
				//Loop for each set on the activity output set
				for(int k = 0; k < outputset_size; k++) {
					if(outputset.get(k).contains(src) && outputset.get(k).contains(dst)) {
						HNSubSet ss = outputset.get(k);
						ss.remove(dst);
						new_outputset.add(ss);
					}
					else {
						HNSubSet ss = outputset.get(k);
						new_outputset.add(ss);
					}
				}
			}
			net.setOutputSet(i, new_outputset);
			net.setInputSet(i, new_inputset);
		}
		
		return net;
	}
	
	public static boolean isSESE(HeuristicsNet net, int[] subnet) {
		int in = -1;
		int out = -1;
		for(int i : subnet) {
			//We have to make sure that the output set of every activity on the subnet is either an element of the subnet of a single element outside the subnet.
			HNSet outputset = net.getOutputSet(i);
			HNSet inputset = net.getInputSet(i);
			
			for(int j = 0; j < outputset.size(); j++) {
				HNSubSet ss = outputset.get(j);
				for(int k = 0; k < ss.size(); k++) {
					if (!integerArrayContains(subnet, ss.get(k))) {
						//The output activity is not present in the activity array...
						if(out == -1 || out == i) { //...but it could be the first one
							out = i;
						}
						else { //...if it's not the first one, the subnet is not a SESE
							return false;
						}
					}
				}
			}
			
			for(int j = 0; j < inputset.size(); j++) {
				HNSubSet ss = inputset.get(j);
				for(int k = 0; k < ss.size(); k++) {
					if (!integerArrayContains(subnet, ss.get(k))) {
						//The input activity is not present in the activity array...
						if(in == -1 || in == i) { //...but it could be the first one
							in = i;
						}
						else { //...if it's not the first one, the subnet is not a SESE
							return false;
						}
					}
				}
			}
			
		}
		return true;
	}

	public static int getInput(HeuristicsNet net, int[] subnet) {
		//You need to make sure the subnet is a SESE before using this method!
		//If it's not a SESE it will return the first activity found outside the subnet.
		for(int i : subnet) {
			HNSet inputset = net.getInputSet(i);
			
			for(int j = 0; j < inputset.size(); j++) {
				HNSubSet ss = inputset.get(j);
				for(int k = 0; k < ss.size(); k++) {
					if (!integerArrayContains(subnet, ss.get(k))) {
						return i;
					}
				}
			}
			
		}
		return -1;
	}
	
	public static int getOutput(HeuristicsNet net, int[] subnet) {
		//You need to make sure the subnet is a SESE before using this method!
		//If it's not a SESE it will return the first activity found outside the subnet.
		for(int i : subnet) {
			HNSet outputset = net.getOutputSet(i);
			
			for(int j = 0; j < outputset.size(); j++) {
				HNSubSet ss = outputset.get(j);
				for(int k = 0; k < ss.size(); k++) {
					if (!integerArrayContains(subnet, ss.get(k))) {
						return i;
					}
				}
			}
		}
		return -1;
	}
	
	private static boolean integerArrayContains(int[] array, int value) {
		for(int i : array) {
			if(i == value) {
				return true;
			}
		}
		return false;
	}
	
	public static HeuristicsNet GroupSubnet(HeuristicsNet net, int[] subnet) {
		//We should group selected subnet making sure it's a SESE
		if (isSESE(net, subnet)) {
			//We erase all activities except one, changing it's name and renaming all variables...
			int input = getInput(net, subnet);
			int output = getOutput(net, subnet);
			
			if(input == -1 && output == -1) {
				//The SESE contains the whole net, so we just return the actual net...
				return net;
			}
			
			HNSet in = (input != -1) ? net.getInputSet(input) : new HNSet();
			HNSet out = (output != -1) ? net.getOutputSet(output) : new HNSet();
			
			for(int i = 0; i < in.size(); i++) {
				HNSubSet ss = in.get(i);
				for(int k : subnet) {
					if(ss.contains(k)) {
						ss.remove(k);
					}
				}
			}
			
			for(int i = 0; i < out.size(); i++) {
				HNSubSet ss = out.get(i);
				for(int k : subnet) {
					if(ss.contains(k)) {
						ss.remove(k);
					}
				}
			}
			
			Map<XEventClass, HNSubSet> activityMapping = net.getActivitiesMappingStructures().getReverseActivitiesMapping();
			for (XEventClass eventClass : activityMapping.keySet()) {
				int id = activityMapping.get(eventClass).get(0);
				if(integerArrayContains(subnet, id)) {
					break;
				}
				HNSet inp = net.getInputSet(id);
//				HNSet outp = net.getOutputSet(id);
				
				for(int i = 0; i < inp.size(); i++) {
					HNSubSet ss = inp.get(i);
					if(ss.contains(output)) {
						ss.remove(output);
						ss.add(input);
					}
				}
//				for(int i = 0; i < outp.size(); i++) {
//					HNSubSet ss = outp.get(i);
//					if(ss.contains(input)) {
//						ss.remove(input);
//						ss.add(input);
//					}
//				}
				
			}
			
			for(int i : subnet) {
				net = (i != input) ? isolateActivity(net, i) : net;
			}
			
			net.setInputSet(input, in);
			net.setOutputSet(input, out);
			
			return net;
		}
		else {
			//If it's not a SESE, we return the same old net, without any changes.
			return net;
		}
	}
	
	public static HeuristicsNet isolateActivity(HeuristicsNet net, int index) {
		//Eliminates all input and output values from other activities to the index activity
		for(int i = 0; i < net.size(); i++) {
			HNSet inputSet = net.getInputSet(i);
			HNSet outputSet = net.getOutputSet(i);
			
			HNSet sinput = new HNSet();
			for(int j = 0; j < inputSet.size(); j++) {
				HNSubSet ss = inputSet.get(j);
				if(ss.contains(index)) {
					//Subset contains index, so we need to erase it from the input set.
					//inputSet.remove(ss);
					ss.remove(index);
				}
				sinput.add(ss);
			}
			net.setInputSet(i, sinput);
			
			HNSet soutput = new HNSet();
			for(int k = 0; k < outputSet.size(); k++) {
				HNSubSet ss = outputSet.get(k);
				if(ss.contains(index)) {
					//Subset contains index, so we need to erase it from the input set.
					//outputSet.remove(ss);
					ss.remove(index);
				}
				soutput.add(ss);
			}
			net.setOutputSet(i, soutput);
		}
		
		net.setInputSet(index, new HNSet());
		net.setOutputSet(index, new HNSet());
		
		return net;
	}
	
	public static Map<RPSTNode, List<Integer>> netProcessing(HeuristicsNet net) {
		
		//Directed graph
		org.jbpt.graph.DirectedGraph graph = new org.jbpt.graph.DirectedGraph();
		
		int size = net.getActivitiesMappingStructures().getXEventClasses().size();
		
		//Loop for each activity on the net
		Vertex vertices[] = new org.jbpt.hypergraph.abs.Vertex[size];
		
		for(int i = 0; i < size; i++) {
			//vertices[i] = new Vertex(net.getActivitiesMappingStructures().getXEventClasses().getByIndex(i).toString());
			vertices[i] = new Vertex(Integer.toString(i));
			graph.addVertex(vertices[i]);
		}
		
		int edges = 0;
		for(int i = 0; i < size; i++) {
			HNSet outputset = net.getOutputSet(i);
			for(int j = 0; j < outputset.size(); j++) {
				HNSubSet outss = outputset.get(j);
				for(int k = 0; k < outss.size(); k++) {
					graph.addEdge(vertices[i],vertices[outss.get(k)]);
					edges++;
				}
			}
		}
		
		System.out.println("New graph:");
		System.out.println("Nodes: " + size);
		System.out.println("Edges: " + edges);
		
		System.out.println(graph);
		RPST rpst = new RPST(graph);
//		System.out.println(rpst.getRPSTNodes());
//		System.out.println("--------------------");
//		
//		Iterator itr = rpst.getRPSTNodes().iterator();
//		while(itr.hasNext()) {			
//			RPSTNode element = (RPSTNode) itr.next();
//	        //System.out.println(element.getName() + ": (" + element.getEntry().toString() + "," + element.getExit().toString() + ") - " + element.getFragment().toString());
//			System.out.print(element.getName() + ": (" + element.getEntry().toString() + "," + element.getExit().toString() + ") - [");
//			Iterator iter = element.getFragment().iterator();			
//	        while(iter.hasNext()) {
//	        	DirectedEdge el = (DirectedEdge) iter.next();
//	        	String output = el.getSource() + "->" + el.getTarget();
//	        	output = iter.hasNext() ? output + ", " : output;
//	        	System.out.print(output);
//	        }
//	        System.out.print("]");
//	        System.out.println();
//	    }
//		
//		System.out.println("--------------------");
		Map<RPSTNode, List<RPSTNode>> m = processRPSTArray(rpst);
		printMappingRPST(m, processRPSTOnNodes(m, graph));
		
		try {
			IOUtils.invokeDOT("/home/npizarro/workspace/", "graph.png", graph.toDOT());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		try {
			IOUtils.invokeDOT("/home/npizarro/workspace/", "rpst.png", rpst.toDOT());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return processRPSTOnNodes(m, graph);
	}
	
	private static Map<RPSTNode, List<RPSTNode>> processRPSTArray(RPST rpst) {
		Map<RPSTNode, List<RPSTNode>> mapping = new HashMap<RPSTNode, List<RPSTNode>>();
		Collection<RPSTNode> rpstNodes = rpst.getRPSTNodes();
		for(RPSTNode v : rpstNodes) {
			List<RPSTNode> l = new ArrayList<RPSTNode>();
			Collection<RPSTNode> children = rpst.getChildren(v);
			if(children.isEmpty()) {
//				mapping.put(v.getName(), l);
			}
			else {
//				for(IVertex ve : children) {
//					l.add(ve);
//				}
				l.addAll(children);
				mapping.put(v, l);
			}
		}
		return mapping;
	}
	
	private static Map<RPSTNode, List<Integer>> processRPSTOnNodes(Map<RPSTNode, List<RPSTNode>> mapping, org.jbpt.graph.DirectedGraph graph) {
		Map<RPSTNode, List<Integer>> map = new HashMap<RPSTNode, List<Integer>>();
		for(RPSTNode n : mapping.keySet()) {
			List<RPSTNode> list = mapping.get(n);
			List<Integer> l = new ArrayList<Integer>();
			List<Vertex> graphVertex = new ArrayList<Vertex>();
			for(RPSTNode no : list) {
				Iterator itr = no.getFragment().iterator();
	            while(itr.hasNext()) {
	            	DirectedEdge edge = (DirectedEdge) itr.next();
	            	if(!graphVertex.contains(edge.getSource())) {
	            		graphVertex.add(edge.getSource());
	            	}
	            	if(!graphVertex.contains(edge.getTarget())) {
	            		graphVertex.add(edge.getTarget());
	            	}
	            }
			}
			for(Vertex v : graphVertex) {
				Collection<Vertex> successors = graph.getDirectSuccessors(v);
	            Collection<Vertex> predecessors = graph.getDirectPredecessors(v);
	            
	            if(graphVertex.containsAll(successors) && graphVertex.containsAll(predecessors)) {
	            	l.add(Integer.parseInt(v.getName()));
	            }
			}
			map.put(n, l);
		}
		return map;
	}
	
	private static void printMappingRPST(Map<RPSTNode, List<RPSTNode>> mapping, Map<RPSTNode, List<Integer>> mappingOnNodes) {
		for(RPSTNode key : mapping.keySet()) {
			List<RPSTNode> list = mapping.get(key);
			List<Integer> l = mappingOnNodes.get(key);
			System.out.println("Node: " + key.getName());
			System.out.print("Children Edges: ");
			for(RPSTNode v : list) {
				System.out.print(v.getName() + " ");
			}
			System.out.println();
			System.out.print("Children Nodes: ");
			for(Integer i : l) {
				System.out.print(i + " ");
			}
			System.out.println();
		}
	}
	
	public static int[] getSESE(Map<RPSTNode, List<Integer>> mappingOnNodes, int index, HeuristicsNet net) {
		List<Integer> res = null;
		int size = Integer.MAX_VALUE;
		for(RPSTNode key : mappingOnNodes.keySet()) {
			List<Integer> list = mappingOnNodes.get(key);
			if(list.contains(index) && list.size() < size && list.size() != 1 && isSESE(net, parseFromInteger(list.toArray(new Integer[list.size()])))) {
				size = list.size();
				res = list;
			}
		}
//		while(!isSESE(net, parseFromInteger(res.toArray(new Integer[res.size()])))) {
//			res.add(getInput(net, parseFromInteger(res.toArray(new Integer[res.size()]))));
//			res.add(getOutput(net, parseFromInteger(res.toArray(new Integer[res.size()]))));
//		}
		return parseFromInteger(res.toArray(new Integer[res.size()]));
	}
	
	public static int[] parseFromInteger(Integer[] arr) {
		int[] res = new int[arr.length];
		for(int i = 0; i < arr.length; i++) {
			res[i] = arr[i];
		}
		return res;
	}
	
	public static HeuristicsNet addEdge(HeuristicsNet net, int src, int dst) {
		//Just for testing purposes, do not use!
		HNSubSet hn_subset = new HNSubSet();
		hn_subset.add(src);
		HNSet hn_set = net.getInputSet(dst);
		hn_set.add(hn_subset);
		net.setInputSet(dst, hn_set);
		
		HNSubSet hn_subset2 = new HNSubSet();
		hn_subset2.add(dst);
		HNSet hn_set2 = net.getOutputSet(src);
		hn_set2.add(hn_subset2);
		net.setOutputSet(src, hn_set2);
		return net;
	}
	
	public static HeuristicsNet eraseEdge(HeuristicsNet net, int src, int dst) {
		//Just for testing purposes, do not use!
		HNSet hn_set = net.getInputSet(dst);
		HNSet new_hn_set = new HNSet();
		for(int i = 0; i < hn_set.size(); i++) {
			HNSubSet hn_subset = hn_set.get(i);
			if(hn_subset.contains(src)) {
				hn_subset.remove(src);
			}
			new_hn_set.add(hn_subset);
		}
		
		net.setInputSet(dst, new_hn_set);
	
		hn_set = net.getOutputSet(src);
		new_hn_set = new HNSet();
		for(int i = 0; i < hn_set.size(); i++) {
			HNSubSet hn_subset = hn_set.get(i);
			if(hn_subset.contains(dst)) {
				hn_subset.remove(dst);
			}
			new_hn_set.add(hn_subset);
		}
		
		net.setOutputSet(src, new_hn_set);
		return net;
	}
	
	public static void logger(long diff, String operation, String source) {
		String text = diff + ";" + operation + ";" + source + "\n";
		Path filePath = Paths.get("/home/npizarro/workspace/operation_log.csv");
		if (!Files.exists(filePath)) {
		    try {
				Files.createFile(filePath);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		try {
			Files.write(filePath, text.getBytes(), StandardOpenOption.APPEND);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	// Reformatted from now on...
	
	public static FeedbackRepairHNVisualization getVisualizationPanel(
			DirectedGraph<?, ?> graph, 
			HeuristicsNet net,
			Progress progress) {
		
		return getResultsPanel(graph, new ViewSpecificAttributeMap(), net, progress);
	}

	public static FeedbackRepairHNVisualization getResultsPanel(DirectedGraph<?, ?> graph,
			ViewSpecificAttributeMap map, HeuristicsNet net, Progress progress) {

		FeedbackRepairProMJGraph jgraph = createJGraph(graph, map, progress);
		
		return new FeedbackRepairHNVisualization(jgraph, net, log);
	}
	
	public static FeedbackRepairProMJGraph createJGraph(DirectedGraph<?, ?> causalNet,
			ViewSpecificAttributeMap map, Progress progress){
		
		GraphLayoutConnection layoutConnection = new GraphLayoutConnection(causalNet);
		
		ProMGraphModel model = new ProMGraphModel(causalNet);
		FeedbackRepairProMJGraph jGraph = new FeedbackRepairProMJGraph(model, map, layoutConnection);
		
		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();
		layout.setDeterministic(false);
		layout.setCompactLayout(false);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(15);
		layout.setFixRoots(false);

		layout.setOrientation(map.get(causalNet, AttributeMap.PREF_ORIENTATION, SwingConstants.SOUTH));

		if(!layoutConnection.isLayedOut()){
		
			JGraphFacade facade = new JGraphFacade(jGraph);
	
			facade.setOrdered(false);
			facade.setEdgePromotion(true);
			facade.setIgnoresCellsInGroups(false);
			facade.setIgnoresHiddenCells(false);
			facade.setIgnoresUnconnectedCells(false);
			facade.setDirected(true);
			facade.resetControlPoints();
			if (layout instanceof JGraphHierarchicalLayout) {
				facade.run((JGraphHierarchicalLayout) layout, true);
			} else {
				facade.run(layout, true);
			}
	
			java.util.Map<?, ?> nested = facade.createNestedMap(true, true);
	
			jGraph.getGraphLayoutCache().edit(nested);
			layoutConnection.setLayedOut(true);
		}
		
		jGraph.setUpdateLayout(layout);
		
		return jGraph;
	}

	protected static JGraphHierarchicalLayout getHierarchicalLayout(final Progress progress) {
		JGraphHierarchicalLayout layout = new JGraphHierarchicalLayout();

		if (progress != null) {
			layout.getProgress().addPropertyChangeListener(new PropertyChangeListener() {

				public void propertyChange(PropertyChangeEvent evt) {
					if (evt.getPropertyName().equals(JGraphLayoutProgress.MAXIMUM_PROPERTY)) {
						progress.setIndeterminate(false);
						progress.setMaximum(((Integer) evt.getNewValue()).intValue());
						Thread.yield();
					} else if (evt.getPropertyName().equals(JGraphLayoutProgress.PROGRESS_PROPERTY)) {
						progress.setValue(((Integer) evt.getNewValue()).intValue());
						Thread.yield();
					}
				}
			});
		}
		layout.setDeterministic(false);
		layout.setCompactLayout(false);
		layout.setFineTuning(true);
		layout.setParallelEdgeSpacing(20);

		return layout;
	}

	@SuppressWarnings("unchecked")
	protected static void fixParallelEdges(JGraphFacade facade, double spacing) {
		ArrayList edges = new ArrayList(facade.getEdges());
		for (Object edge : edges) {
			List points = facade.getPoints(edge);
			if (points.size() != 2) {
				continue;
			}
			Object sourceCell = facade.getSource(edge);
			Object targetCell = facade.getTarget(edge);
			Object sourcePort = facade.getSourcePort(edge);
			Object targetPort = facade.getTargetPort(edge);
			Object[] between = facade.getEdgesBetween(sourcePort, targetPort, false);
			if ((between.length == 1) && !(sourcePort == targetPort)) {
				continue;
			}
			Rectangle2D sCP = facade.getBounds(sourceCell);
			Rectangle2D tCP = facade.getBounds(targetCell);
			Point2D sPP = GraphConstants.getOffset(((ProMGraphPort) sourcePort).getAttributes());

			if (sPP == null) {
				sPP = new Point2D.Double(sCP.getCenterX(), sCP.getCenterY());
			}
			Point2D tPP = GraphConstants.getOffset(((ProMGraphPort) targetPort).getAttributes());

			if (tPP == null) {
				tPP = new Point2D.Double(tCP.getCenterX(), tCP.getCenterY());
			}

			if (sourcePort == targetPort) {
				assert (sPP.equals(tPP));
				double x = sPP.getX();
				double y = sPP.getY();
				for (int i = 2; i < between.length + 2; i++) {
					List newPoints = new ArrayList(5);
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y));
					newPoints.add(new Point2D.Double(x - (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x, y - (2 * spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing + i * spacing), y - (spacing + i * spacing)));
					newPoints.add(new Point2D.Double(x + (spacing), y - (spacing / 2 + i * spacing)));
					facade.setPoints(between[i - 2], newPoints);
				}

				continue;
			}

			double dx = (sPP.getX()) - (tPP.getX());
			double dy = (sPP.getY()) - (tPP.getY());
			double mx = (tPP.getX()) + dx / 2.0;
			double my = (tPP.getY()) + dy / 2.0;
			double slope = Math.sqrt(dx * dx + dy * dy);
			for (int i = 0; i < between.length; i++) {
				List newPoints = new ArrayList(3);
				double pos = 2 * i - (between.length - 1);
				if (facade.getSourcePort(between[i]) == sourcePort) {
					newPoints.add(sPP);
					newPoints.add(tPP);
				} else {
					newPoints.add(tPP);
					newPoints.add(sPP);
				}
				if (pos != 0) {
					pos = pos / 2;
					double x = mx + pos * spacing * dy / slope;
					double y = my - pos * spacing * dx / slope;
					newPoints.add(1, new SerializablePoint2D.Double(x, y));
				}
				facade.setPoints(between[i], newPoints);
			}
		}
	}

	public static JComponent visualizeGraph(
			DirectedGraph<?, ?> graph, 
			HeuristicsNet net,
			Progress progress) {
		
		return getResultsPanel(graph, new ViewSpecificAttributeMap(), net, progress);
	}
}
