package com.processmininguc.feedbackrepair.models;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.jbpt.algo.tree.rpst.RPSTNode;
import org.jgraph.JGraph;
import org.jgraph.event.GraphSelectionEvent;
import org.jgraph.event.GraphSelectionListener;
import org.jgraph.graph.BasicMarqueeHandler;
import org.processmining.framework.util.Cleanable;
import org.processmining.framework.util.ui.scalableview.VerticalLabelUI;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.ViewSpecificAttributeMap;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.models.heuristics.HeuristicsNetGraph;
import org.processmining.models.heuristics.impl.ActivitiesMappingStructures;
import org.processmining.models.heuristics.impl.HNSet;
import org.processmining.models.heuristics.impl.HNSubSet;
import org.processmining.models.jgraph.ContextMenuCreator;
import org.processmining.models.jgraph.ProMGraphModel;
import org.processmining.models.jgraph.elements.ProMGraphCell;
import org.processmining.models.jgraph.elements.ProMGraphEdge;
import org.processmining.models.jgraph.listeners.SelectionListener;
import org.processmining.plugins.heuristicsnet.SimpleHeuristicsNet;
import org.processmining.plugins.heuristicsnet.miner.heuristics.miner.fitness.ImprovedContinuousSemantics;

import com.fluxicon.slickerbox.factory.SlickerDecorator;
import com.fluxicon.slickerbox.factory.SlickerFactory;

public class FeedbackRepairHNVisualization extends JPanel implements Cleanable,
ChangeListener, ActionListener  {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8100193630289342766L;

	protected static final int MAX_ZOOM = 1200;

	// -------------------------------------------

	protected FeedbackRepairProMJGraph graph;
	protected HeuristicsNet net;
	protected HeuristicsNet net_bkp;
	
	private Map<RPSTNode, List<Integer>> RPSTmappingOnNodes;

	private JPanel buttonHolder;
	private ProMGraphCell src;
	private ProMGraphCell dst;
	private ArrayList<ProMGraphCell> selectedCells;
	private boolean firstSelection;
	Map<XEventClass, HNSubSet> activityMapping;
	Map<String, Integer> reverseActivityMapping;
	Map<Integer, String> customActivityMapping;
	private boolean flagSetSource;
	private FeedbackRepairOperation op;
	private XLog log;
	private XLogInfo logInfo;
	private String textSESE;
	// -------------------------------------------

	protected JScrollPane scroll;
	private JPanel zoomPanelON, zoomPanelOFF;
	private ZoomPanel zoom;
	private JPanel parametersPanelON, parametersPanelOFF;
	private CnetInputOutput parameters;
	private JPanel fitnessPanel;

	private float zoomRatio;
	private double normalScale;
	private Rectangle normalBounds, zoomBounds;

	private boolean hasNodeSelected;

	private List<SelectionListener<?, ?>> selectionListeners = new ArrayList<SelectionListener<?, ?>>(
			0);
	private ContextMenuCreator creator = null;

	public FeedbackRepairHNVisualization(final FeedbackRepairProMJGraph sgraph,
			final HeuristicsNet snet, XLog slog) {
		
		this.setLayout(null);
		this.graph = sgraph;
		this.net = snet;
		this.net_bkp = this.net.copy();
		this.RPSTmappingOnNodes = FeedbackRepair.netProcessing(this.net);
		this.selectedCells = new ArrayList<ProMGraphCell>();
		this.firstSelection = true;
		this.activityMapping = this.net.getActivitiesMappingStructures().getReverseActivitiesMapping();
		this.reverseActivityMapping = new HashMap<String, Integer>();
		this.customActivityMapping = new HashMap<Integer, String>();
		for (XEventClass eventClass : activityMapping.keySet()) {
			reverseActivityMapping.put(eventClass.toString(), activityMapping.get(eventClass).get(0));
			customActivityMapping.put(activityMapping.get(eventClass).get(0), eventClass.toString());
		}
		this.flagSetSource = false;
		this.log = slog;
		this.logInfo = XLogInfoFactory.createLogInfo(this.log);

		SlickerFactory factory = SlickerFactory.instance();
		SlickerDecorator decorator = SlickerDecorator.instance();

		this.addComponentListener(new java.awt.event.ComponentListener() {

			public void componentHidden(ComponentEvent e) {
			}

			public void componentMoved(ComponentEvent e) {
			}

			public void componentShown(ComponentEvent e) {
			}

			public void componentResized(ComponentEvent e) {
				resize();
			}
		});

		this.initGraph();

		AdjustmentListener aListener = new AdjustmentListener() {
			public void adjustmentValueChanged(AdjustmentEvent e) {
				
			}
		};
		MouseListener mListener = new MouseListener() {

			public void mouseClicked(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {

				if (hasNodeSelected) {
					
				}
			}

			public void mouseExited(MouseEvent e) {

				if (hasNodeSelected) {

					scroll.repaint();
				}

			}

			public void mousePressed(MouseEvent e) {

				zoomPanelOFF.setVisible(false);
				zoomPanelOFF.setEnabled(false);

				parametersPanelOFF.setVisible(false);
				parametersPanelOFF.setEnabled(false);
			}

			public void mouseReleased(MouseEvent e) {

				zoomPanelOFF.setVisible(true);
				zoomPanelOFF.setEnabled(true);

				parametersPanelOFF.setVisible(true);
				parametersPanelOFF.setEnabled(true);
			}
		};

		this.scroll = new JScrollPane(graph);
		this.scroll.getHorizontalScrollBar().addAdjustmentListener(aListener);
		this.scroll.getVerticalScrollBar().addAdjustmentListener(aListener);
		this.scroll.getHorizontalScrollBar().addMouseListener(mListener);
		this.scroll.getVerticalScrollBar().addMouseListener(mListener);
		decorator.decorate(this.scroll, Color.WHITE, Color.GRAY, Color.DARK_GRAY);
		this.scroll.setBorder(new DashedBorder(Color.LIGHT_GRAY));

		this.zoomPanelON = factory.createRoundedPanel(15, Color.LIGHT_GRAY);
		this.zoomPanelOFF = factory.createRoundedPanel(15, Color.DARK_GRAY);
		this.zoomPanelON.setLayout(null);
		this.zoomPanelOFF.setLayout(null);
		this.zoom = new ZoomPanel(factory, decorator, 100, MAX_ZOOM);
		this.zoom.addSliderChangeListener(this);
		this.zoomPanelON.add(this.zoom);
		this.zoomPanelON.setVisible(false);
		this.zoomPanelON.setEnabled(false);
		JLabel zoomPanelTitle = factory.createLabel("Zoom");
		zoomPanelTitle.setForeground(Color.WHITE);
		zoomPanelTitle.setFont(new java.awt.Font("Dialog", java.awt.Font.BOLD, 18));
		zoomPanelTitle.setUI(new VerticalLabelUI(true));
		this.zoomPanelOFF.add(zoomPanelTitle);
		zoomPanelTitle.setBounds(10, 10, 30, 55);

		this.zoomPanelOFF.addMouseListener(new MouseListener() {

			public void mouseClicked(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {

				int x = scroll.getHorizontalScrollBar().getValue();
				int y = scroll.getVerticalScrollBar().getValue();

				showParameters(false);
				showZoom(true);

				scroll.getHorizontalScrollBar().setValue((int) (x * zoomRatio));
				scroll.getVerticalScrollBar().setValue((int) (y * zoomRatio));

				if (hasNodeSelected) {
					
				}
			}

			public void mouseExited(MouseEvent e) {
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {
			}
		});

		this.zoomPanelON.addMouseListener(new MouseListener() {

			public void mouseClicked(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {
			}

			public void mouseExited(MouseEvent e) {

				int x = scroll.getHorizontalScrollBar().getValue();
				int y = scroll.getVerticalScrollBar().getValue();

				boolean flag = false;
				if (e.getX() >= zoom.getWidth())
					flag = true;
				else {

					if ((e.getY() >= zoom.getHeight()) || (e.getY() <= 0))
						flag = true;
				}

				if (flag) {

					showZoom(false);

					scroll.getHorizontalScrollBar().setValue(
							(int) (x / zoomRatio));
					scroll.getVerticalScrollBar().setValue(
							(int) (y / zoomRatio));

					if (hasNodeSelected) {
						
					}
				}
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {
			}
		});

		if (this.net instanceof SimpleHeuristicsNet) {

			this.parametersPanelON = factory.createRoundedPanel(15,
					Color.LIGHT_GRAY);
			this.parametersPanelOFF = factory.createRoundedPanel(15,
					Color.DARK_GRAY);
			this.parametersPanelON.setLayout(null);
			this.parametersPanelOFF.setLayout(null);
			this.parameters = new CnetInputOutput(this.net, this.textSESE);
			this.parameters.setBounds(10, 10, 520, 390);
			this.parameters.setBackground(Color.LIGHT_GRAY);
			this.parameters.setEnabled(false);
			this.parametersPanelON.add(this.parameters);
			this.parametersPanelON.setVisible(false);
			this.parametersPanelON.setEnabled(false);
			JLabel parametersPanelTitle = factory.createLabel("C-net I/O");
			parametersPanelTitle.setForeground(Color.WHITE);
			parametersPanelTitle.setFont(new java.awt.Font("Dialog",
					java.awt.Font.BOLD, 18));
			this.parametersPanelOFF.add(parametersPanelTitle);
			parametersPanelTitle.setBounds(12, 0, 105, 30);

			this.parametersPanelOFF.addMouseListener(new MouseListener() {

				public void mouseClicked(MouseEvent e) {
				}

				public void mouseEntered(MouseEvent e) {

					showZoom(false);
					showParameters(true);

					if (hasNodeSelected) {
						
					}
				}

				public void mouseExited(MouseEvent e) {
				}

				public void mousePressed(MouseEvent e) {
				}

				public void mouseReleased(MouseEvent e) {
				}
			});

			this.parametersPanelON.addMouseListener(new MouseListener() {

				public void mouseClicked(MouseEvent e) {
				}

				public void mouseEntered(MouseEvent e) {
				}

				public void mouseExited(MouseEvent e) {

					boolean flag = false;
					if ((e.getX() >= parametersPanelON.getWidth())
							|| (e.getX() <= 0))
						flag = true;
					else {

						if ((e.getY() >= parametersPanelON.getHeight())
								|| (e.getY() <= 0))
							flag = true;
					}

					if (flag) {

						showParameters(false);

						if (hasNodeSelected) {

						}
					}
				}

				public void mousePressed(MouseEvent e) {
				}

				public void mouseReleased(MouseEvent e) {
				}
			});
		} else
			this.parametersPanelON = new JPanel();
		
		//Panel de botones
		JButton b1 = factory.createButton("->");
		JButton b2 = factory.createButton("||");
		JButton b3 = factory.createButton("#");
		JButton b4 = factory.createButton("Update");
		JButton b5 = factory.createButton("Restore");
		JButton b6 = factory.createButton("Group");
		JButton b7 = factory.createButton("RPST");
	
		this.buttonHolder = factory.createRoundedPanel(15, Color.LIGHT_GRAY);
		this.buttonHolder.setLayout(null);
		this.buttonHolder.add(b1);
		this.buttonHolder.add(b2);
		this.buttonHolder.add(b3);
		this.buttonHolder.add(b4);
		this.buttonHolder.add(b5);
		this.buttonHolder.add(b6);
		this.buttonHolder.add(b7);
		
		b1.setBounds(0, 0, 100, 30);
		b2.setBounds(100, 0, 100, 30);
		b3.setBounds(200, 0, 100, 30);
		b4.setBounds(300, 0, 100, 30);
		b5.setBounds(400, 0, 100, 30);
		b6.setBounds(500, 0, 100, 30);
		b7.setBounds(600, 0, 100, 30);
		
		b1.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Click ->");
				op = FeedbackRepairOperation.FOLLOWS;
				flagSetSource = true;
			}
		});
		
		b2.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Click ||");
				op = FeedbackRepairOperation.PARALLEL;
				flagSetSource = true;
			}
		});
		
		b3.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Click #");
				op = FeedbackRepairOperation.INDIFFERENT;
				flagSetSource = true;
			}
		});
		
		b4.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Updating...");
				showParameters(true);
				
				
//				activityMapping = net.getActivitiesMappingStructures().getReverseActivitiesMapping();
//				reverseActivityMapping = new HashMap<String, Integer>();
				
//				XEventClasses c = net.getActivitiesMappingStructures().getXEventClasses();
//				Collection<XEventClass> xec = c.getClasses();
//				for(XEventClass cl : xec) {
//					if(cl.toString() == "Alguna actividad") {
//						xec.remove(cl);
//					}
//				}
				
//				XEventClassifier ec = c.getClassifier();
//				XEventClasses c_aux = new XEventClasses(ec);
				
//				for (XEventClass eventClass : activityMapping.keySet()) {
//					reverseActivityMapping.put(eventClass.toString(), activityMapping.get(eventClass).get(0));
//				}
				long start = System.nanoTime();
				flagSetSource = false;
				net = FeedbackRepair.ModifyNet(net, op, reverseActivityMapping.get(src.getNode().getLabel()), reverseActivityMapping.get(dst.getNode().getLabel()));
				RPSTmappingOnNodes = FeedbackRepair.netProcessing(net);
				redraw();
				long end = System.nanoTime();
				FeedbackRepair.logger(end - start, op.toString(), "NETMOD + VIS");
			}
		});
		
		b5.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Restoring...");
				showParameters(true);
				
				flagSetSource = false;
				net = net_bkp.copy();
				RPSTmappingOnNodes = FeedbackRepair.netProcessing(net);
				redraw();
			}
		});
		
		b6.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Trying to group selected activities. SESE status: " + FeedbackRepair.isSESE(net, ProMCellToIntegerArray(selectedCells)));
				showParameters(true);
				
				if (FeedbackRepair.isSESE(net, ProMCellToIntegerArray(selectedCells))) {
					//Subnet is a SESE, we can group.
					net = FeedbackRepair.GroupSubnet(net, ProMCellToIntegerArray(selectedCells));
					//We generate the new activity mapping structure...
					
//					ActivitiesMappingStructures ams = hideActivities(ProMCellToIntegerArray(selectedCells));
//					net.setActivitiesMappingStructures(ams);
//					
//					activityMapping = net.getActivitiesMappingStructures().getReverseActivitiesMapping();
//					for (XEventClass eventClass : activityMapping.keySet()) {
//						reverseActivityMapping.put(eventClass.toString(), activityMapping.get(eventClass).get(0));
//						customActivityMapping.put(activityMapping.get(eventClass).get(0), eventClass.toString());
//					}
					
					redraw();
				}
			}
		});
		
		b7.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent arg0) {
				System.out.println("Processing for a RPST...");
				showParameters(false);
				
				//Process the net for an RPST
				RPSTmappingOnNodes = FeedbackRepair.netProcessing(net);
				redraw();
			}
		});
		
		updateCustomListener();

		this.fitnessPanel = factory.createRoundedPanel(15, Color.GRAY);
		this.fitnessPanel.setLayout(null);
		JLabel fitnessInfo = factory.createLabel("Fitness: "
				+ (Math.round(this.net.getFitness() * 10000) / 10000f));
		fitnessInfo.setForeground(Color.WHITE);
		fitnessInfo
				.setFont(new java.awt.Font("Dialog", java.awt.Font.PLAIN, 18));
		this.fitnessPanel.add(fitnessInfo);
		fitnessInfo.setBounds(20, 0, 140, 30);

		this.add(this.zoomPanelON);
		this.add(this.zoomPanelOFF);
		this.add(this.parametersPanelON);
		this.add(this.parametersPanelOFF);
		this.add(this.fitnessPanel);
		this.add(this.scroll);
		this.add(this.buttonHolder);

		this.setBackground(Color.WHITE);

		this.validate();
		this.repaint();
	}
	
	private ActivitiesMappingStructures hideActivities(int[] hiddenActivities) {
		XEventClasses c = net.getActivitiesMappingStructures().getXEventClasses();
		Map<XEventClass, Integer> mapping = new HashMap<XEventClass, Integer>();
		for(XEventClass cl : c.getClasses()) {
			mapping.put(cl, 1);
			for(int i = 0; i < hiddenActivities.length; i++) {
				if(hiddenActivities[i] == activityMapping.get(cl).get(0)) {
					mapping.put(cl, 0);
					break;
				}
			}
		}
		return new ActivitiesMappingStructures(c, mapping);	
	}
	
	private void updateCustomListener() {
		this.graph.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				if (e.getSource() instanceof JGraph) {
					final JGraph graph = (JGraph) e.getSource();
					Object cell = graph.getFirstCellForLocation(e.getX(), e.getY());
					if (cell != null && cell instanceof ProMGraphCell) {
					  ProMGraphCell clicked = (ProMGraphCell) cell;
					  if(!flagSetSource) {
						  src = clicked;
						  System.out.println("Source: " + clicked.getNode().getLabel());
					  }
					  else {
						  dst = clicked;
						  System.out.println("Destination: " + clicked.getNode().getLabel());
					  }
					  System.out.println("Maximal SESE:");
					  int[] sese = FeedbackRepair.getSESE(RPSTmappingOnNodes, reverseActivityMapping.get(clicked.getNode().getLabel()), net);
					  int inpt = FeedbackRepair.getInput(net, sese);
					  int outpt = FeedbackRepair.getOutput(net, sese);
					  textSESE = "Identified SESE: (" + sese.length + " activities)\n";
					  for(int i : sese) {
						  System.out.println("    " + i + " [" + customActivityMapping.get(i) + "]");
						  if(i == inpt) {
							  textSESE += i + " [" + customActivityMapping.get(i) + "] <--- SESE Input\n";
						  }
						  else if(i == outpt) {
							  textSESE += i + " [" + customActivityMapping.get(i) + "] <--- SESE Output\n";
						  }
						  else {
							  textSESE += i + " [" + customActivityMapping.get(i) + "]\n";
						  }
					  }
//			          String lab = graph.convertValueToString(cell);
//			          System.out.println(lab);
			        }
				}
				redrawParameters();
			}
		});
		
		this.graph.addGraphSelectionListener(new GraphSelectionListener() {
			public void valueChanged(GraphSelectionEvent gse) {
				//valueChanged returns both current and last selection (cause both of them actually changed, ones entering and others leaving the array).
				// We need to initialize the variables the first time a selection is made, in order to determine which elements are new on upcoming selections.
				System.out.println("Selection has changed!");
				selectedCells.clear();
				for(Object o : gse.getCells()) {
					if(o instanceof ProMGraphCell) {
						ProMGraphCell cell = (ProMGraphCell) o;
						//System.out.println("    " + cell.getNode().getLabel() + " - " + gse.isAddedCell(o));
						if(firstSelection) {
							System.out.println("    " + cell.getNode().getLabel());
							selectedCells.add(cell);
						}
						else {
							if(gse.isAddedCell(o)) {
								System.out.println("    " + cell.getNode().getLabel());
								selectedCells.add(cell);
							}
						}
					}
				}
				firstSelection = false;
				System.out.println("SESE status: " + FeedbackRepair.isSESE(net, ProMCellToIntegerArray(selectedCells)));
			}
		});
	}
	
	private int[] ProMCellToIntegerArray(ArrayList<ProMGraphCell> cells) {
		int[] result = new int[cells.size()];
		
		for(int i = 0; i < cells.size(); i++) {
			result[i] = reverseActivityMapping.get(cells.get(i).getNode().getLabel());
		}
		return result;
	}

	private void resize() {

		int width = this.getSize().width;
		int height = this.getSize().height;

		this.zoom.setHeight((int) (height * 0.66));

		int zoomWidth = this.zoom.getSize().width;
		int zoomHeight = this.zoom.getSize().height;

		this.zoomRatio = (float) (width - zoomWidth - 40)
				/ (float) (width - 60);
		this.normalBounds = new Rectangle(30, 30, width - 60, height - 60);
		this.zoomBounds = new Rectangle(10 + zoomWidth,
				30 + (int) ((1f - this.zoomRatio) * (height - 60)), width
						- zoomWidth - 40,
				(int) (this.zoomRatio * (height - 60)));

		this.normalScale = graph.getScale();

		this.scroll.setBounds(this.normalBounds);
		this.zoomPanelON.setBounds(0, 40, zoomWidth + 10, zoomHeight);
		this.zoomPanelOFF.setBounds(-10, 40, 40, 72);

		int parametersHeight = this.parameters.getHeight();
		this.parametersPanelON.setBounds(width - 580,
				height - parametersHeight, 540, parametersHeight + 10);
		this.parametersPanelOFF.setBounds(width - 165, height - 30, 125, 40);

		this.fitnessPanel.setBounds(-10, height - 30, 160, 40);
		this.buttonHolder.setBounds((width / 2) - 350, height - 30, 700, 40);

		double fitRatio = scaleToFit(this.graph, this.scroll, false);
		this.zoom.setFitValue((int) Math.floor(fitRatio * this.zoomRatio * 100));
	}

	private void showZoom(boolean status) {

		zoomPanelOFF.setVisible(!status);
		zoomPanelOFF.setEnabled(!status);
		zoomPanelON.setVisible(status);
		zoomPanelON.setEnabled(status);

		if (status) {

			this.scroll.setBounds(this.zoomBounds);
			graph.setScale(this.normalScale * this.zoomRatio);
		} else {

			this.scroll.setBounds(this.normalBounds);
			graph.setScale(this.normalScale);
		}

	}

	private void showParameters(boolean status) {

		parametersPanelOFF.setVisible(!status);
		parametersPanelOFF.setEnabled(!status);
		parametersPanelON.setVisible(status);
		parametersPanelON.setEnabled(status);
	}

	private void redraw() {
		int scrollPositionX = this.scroll.getHorizontalScrollBar().getValue();
		int scrollPositionY = this.scroll.getVerticalScrollBar().getValue();

		HeuristicsNetGraph hng = new HeuristicsNetGraph(this.net, "Heuristics Net", false);
		
		this.graph = FeedbackRepair.createJGraph(hng,
				new ViewSpecificAttributeMap(), null);
		
		this.initGraph();

		this.remove(this.scroll);

		AdjustmentListener aListener = new AdjustmentListener() {
			public void adjustmentValueChanged(AdjustmentEvent e) {
				
			}
		};
		MouseListener mListener = new MouseListener() {

			public void mouseClicked(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {

				if (hasNodeSelected) {
					
				}
			}

			public void mouseExited(MouseEvent e) {

				if (hasNodeSelected) {
					
				}
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {
			}
		};

		this.scroll = new JScrollPane(graph);
		SlickerDecorator.instance().decorate(this.scroll, Color.WHITE, Color.GRAY, Color.DARK_GRAY);
		this.scroll.getHorizontalScrollBar().addAdjustmentListener(aListener);
		this.scroll.getVerticalScrollBar().addAdjustmentListener(aListener);
		this.scroll.getHorizontalScrollBar().addMouseListener(mListener);
		this.scroll.getVerticalScrollBar().addMouseListener(mListener);
		this.scroll.setBorder(new DashedBorder(Color.LIGHT_GRAY));
		this.add(this.scroll);
		this.scroll.setBounds(this.normalBounds);

		this.graph.setScale(this.normalScale);

		this.scroll.getHorizontalScrollBar().setValue(scrollPositionX);
		this.scroll.getVerticalScrollBar().setValue(scrollPositionY);
		
		updateCustomListener();
		HeuristicsNet[] population = new HeuristicsNet[1];
		population[0] = net;

//		ContinuousSemantics fitness1 = new ContinuousSemantics(logInfo);
//		System.out.println(fitness1.calculate(population)[0].getFitness());

		ImprovedContinuousSemantics fitness2 = new ImprovedContinuousSemantics(logInfo);
		double fitness = fitness2.calculate(population)[0].getFitness();
		System.out.println("New fitness: " + fitness);
		net.setFitness(fitness);
		
		SlickerFactory factory = SlickerFactory.instance();
		this.fitnessPanel.removeAll();
//		this.fitnessPanel = factory.createRoundedPanel(15, Color.GRAY);
//		this.fitnessPanel.setLayout(null);
		JLabel fitnessInfo = factory.createLabel("Fitness: " + (Math.round(this.net.getFitness() * 10000) / 10000f));
		fitnessInfo.setForeground(Color.WHITE);
		fitnessInfo.setFont(new java.awt.Font("Dialog", java.awt.Font.PLAIN, 18));
		this.fitnessPanel.add(fitnessInfo);
		fitnessInfo.setBounds(20, 0, 140, 30);

		redrawParameters();
	}
	
	private void redrawParameters() {
		this.parametersPanelON.remove(this.parameters);
		this.parameters.removeAll();
		this.parameters = new CnetInputOutput(this.net, this.textSESE);
		this.parameters.setBounds(10, 10, 520, 390);
		this.parameters.setBackground(Color.LIGHT_GRAY);
		this.parameters.setEnabled(false);
		this.parametersPanelON.add(this.parameters);
	}

	private void initGraph() {
		
		this.graph.addGraphSelectionListener(new GraphSelectionListener() {

			@SuppressWarnings("unchecked")
			public void valueChanged(GraphSelectionEvent e) {

				DirectedGraphNode selectedCell = null;

				Object[] cells = e.getCells();
				Collection nodesAdded = new ArrayList<ProMGraphCell>();
				Collection edgesAdded = new ArrayList<ProMGraphEdge>();
				Collection nodesRemoved = new ArrayList<ProMGraphCell>();
				Collection edgesRemoved = new ArrayList<ProMGraphEdge>();
				Collection<?> nodes = graph.getProMGraph().getNodes();
				Collection<?> edges = graph.getProMGraph().getEdges();
				for (int i = 0; i < cells.length; i++) {
					Collection nodeList;
					Collection edgeList;

					boolean isCell = cells[i] instanceof ProMGraphCell;
					boolean isEdge = cells[i] instanceof ProMGraphEdge;

					if (e.isAddedCell(i)) {
						nodeList = nodesAdded;
						edgeList = edgesAdded;

						if (isCell && (selectedCell == null))
							selectedCell = ((ProMGraphCell) cells[i]).getNode();

					} else {
						nodeList = nodesRemoved;
						edgeList = edgesRemoved;
					}
					if (isCell) {
						DirectedGraphNode node = ((ProMGraphCell) cells[i])
								.getNode();
						if (nodes.contains(node)) {
							nodeList.add(node);
						}
					} else if (isEdge) {
						DirectedGraphEdge<?, ?> edge = ((ProMGraphEdge) cells[i])
								.getEdge();
						if (edges.contains(edge)) {
							edgeList.add(((ProMGraphEdge) cells[i]).getEdge());
						}
					}
				}
				SelectionListener.SelectionChangeEvent event = new SelectionListener.SelectionChangeEvent(
						nodesAdded, edgesAdded, nodesRemoved, edgesRemoved);
				for (SelectionListener listener : selectionListeners) {
					listener.SelectionChanged(event);
				}
			}

		});
		this.graph.setTolerance(4);

		this.graph.setMarqueeHandler(new BasicMarqueeHandler() {
			private boolean test(MouseEvent e) {
				return SwingUtilities.isRightMouseButton(e)
						&& (e.getModifiers() & InputEvent.ALT_MASK) == 0;

			}

			public boolean isForceMarqueeEvent(MouseEvent event) {
				if (test(event)) {
					return true;
				} else {
					return false;
				}
			}

			@Override
			public void mouseReleased(final MouseEvent e) {
				if (test(e)) {
					e.consume();
				} else {
					super.mouseReleased(e);
				}
			}

			@Override
			public void mousePressed(final MouseEvent e) {
				if (test(e)) {
					synchronized (graph.getProMGraph()) {
						// Check for selection.
						// If the cell that is being clicked is part of the
						// selection,
						// we use the current selection.
						// otherwise, we use a new selection
						Object cell = graph.getFirstCellForLocation(e.getX(), e
								.getY());

						Collection<DirectedGraphElement> sel;
						if (cell == null) {
							// Nothing selected
							graph.clearSelection();
							sel = new ArrayList<DirectedGraphElement>(0);
						} else if (graph.getSelectionModel().isCellSelected(
								cell)) {
							// the current selection contains cell
							// use that selection
							sel = getSelectedElements();
						} else {
							// the current selection does not contain cell.
							// reset the selection to [cell]
							sel = new ArrayList<DirectedGraphElement>(1);
							sel.add(getElementForLocation(e.getX(), e.getY()));
							graph.setSelectionCell(cell);
						}
						if (creator != null) {
							JPopupMenu menu = creator.createMenuFor(graph
									.getProMGraph(), sel);
							if (menu != null) {
								menu.show(graph, e.getX(), e.getY());
							}
						}
					}
				} else {
					super.mousePressed(e);
				}
			}
		});

		this.graph.addMouseListener(new MouseListener() {

			public void mouseClicked(MouseEvent e) {
			}

			public void mouseEntered(MouseEvent e) {
			}

			public void mouseExited(MouseEvent e) {
			}

			public void mousePressed(MouseEvent e) {
			}

			public void mouseReleased(MouseEvent e) {

				if (hasNodeSelected) {
					
				}
			}
		});

		// Collapse any expandable nodes that claim they are collapsed
		// This is not handled previously.
//		for (Object n : graph.getGraphLayoutCache().getCells(true, false,
//				false, false)) {
//			if (((ProMGraphCell) n).getNode() instanceof Expandable) {
//				Expandable ex = (Expandable) ((ProMGraphCell) n).getNode();
//				
//				if (ex.isCollapsed()) { ex.collapse(); }
//			}
//		}
		
		GraphLayoutConnection con = new GraphLayoutConnection(this.graph.getProMGraph());

		ProMGraphModel model = new ProMGraphModel(graph.getProMGraph());

		this.hasNodeSelected = false;
	}

	public double getScale() {
		return graph.getScale();
	}

	public void setScale(double d) {

		int b = (int) (100.0 * d);
		b = Math.max(b, 1);
		b = Math.min(b, MAX_ZOOM);
		this.zoom.setValue(b);
	}

	protected double scaleToFit(FeedbackRepairProMJGraph graph, Container container,
			boolean reposition) {

		Rectangle2D bounds = graph.getBounds();
		double x = bounds.getX();
		double y = bounds.getY();
		if (reposition) {

			graph.repositionToOrigin();
			x = 0;
			y = 0;
		}

		Dimension size = container.getSize();

		double ratio = Math.min(size.getWidth() / (bounds.getWidth() + x), size
				.getHeight()
				/ (bounds.getHeight() + y));

		return ratio;
	}

	public void paint(Graphics g) {
		super.paint(g);
	}

	public DirectedGraphElement getElementForLocation(double x, double y) {
		Object cell = graph.getFirstCellForLocation(x, y);
		if (cell instanceof ProMGraphCell) {
			return ((ProMGraphCell) cell).getNode();
		}
		if (cell instanceof ProMGraphEdge) {
			return ((ProMGraphEdge) cell).getEdge();
		}
		return null;
	}

	public Collection<DirectedGraphNode> getSelectedNodes() {
		List<DirectedGraphNode> nodes = new ArrayList<DirectedGraphNode>();
		for (Object o : graph.getSelectionCells()) {
			if (o instanceof ProMGraphCell) {
				nodes.add(((ProMGraphCell) o).getNode());
			}
		}
		return nodes;
	}

	public Collection<DirectedGraphEdge<?, ?>> getSelectedEdges() {
		List<DirectedGraphEdge<?, ?>> edges = new ArrayList<DirectedGraphEdge<?, ?>>();
		for (Object o : graph.getSelectionCells()) {
			if (o instanceof ProMGraphEdge) {
				edges.add(((ProMGraphEdge) o).getEdge());
			}
		}
		return edges;
	}

	public Collection<DirectedGraphElement> getSelectedElements() {
		List<DirectedGraphElement> elements = new ArrayList<DirectedGraphElement>();
		for (Object o : graph.getSelectionCells()) {
			if (o instanceof ProMGraphCell) {
				elements.add(((ProMGraphCell) o).getNode());
			} else if (o instanceof ProMGraphEdge) {
				elements.add(((ProMGraphEdge) o).getEdge());
			}
		}
		return elements;
	}

	public void cleanUp() {

		graph.cleanUp();
	}

	public void stateChanged(ChangeEvent e) {

		Object source = e.getSource();

		if (source instanceof JSlider) {

			graph.setScale(((JSlider) source).getValue() / 100.0);

			this.normalScale = graph.getScale() / this.zoomRatio;
		}
	}

	public void actionPerformed(ActionEvent e) {
		// TODO Auto-generated method stub

	}

	@Override
	public void finalize() throws Throwable {

		try {
			cleanUp();
		} finally {
			super.finalize();
		}
	}

}

class ZoomPanel extends JPanel {

	/**
	 * 
	 */
	private static final long serialVersionUID = 8415559591750873766L;
	private final JSlider slider;
	private JLabel sliderMinValue, sliderMaxValue, sliderFitValue, sliderValue;

	private int fitZoom;

	public ZoomPanel(SlickerFactory factory, SlickerDecorator decorator,
			int fitZoom, int maximumZoom) {

		super(null);

		this.fitZoom = fitZoom;

		this.slider = factory.createSlider(1);

		this.slider.setMinimum(1);
		this.slider.setMaximum(maximumZoom);
		this.slider.setValue(fitZoom);

		this.slider.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				update();
			}
		});

		this.sliderMinValue = factory.createLabel("0%");
		this.sliderMaxValue = factory.createLabel(maximumZoom + "%");
		this.sliderFitValue = factory.createLabel("Fit  >");
		this.sliderValue = factory.createLabel(fitZoom + "%");

		this.sliderMinValue.setHorizontalAlignment(SwingConstants.CENTER);
		this.sliderMaxValue.setHorizontalAlignment(SwingConstants.CENTER);
		this.sliderFitValue.setHorizontalAlignment(SwingConstants.RIGHT);
		this.sliderValue.setHorizontalAlignment(SwingConstants.LEFT);

		this.sliderMinValue.setFont(new java.awt.Font("Dialog",
				java.awt.Font.BOLD, 14));
		this.sliderMaxValue.setFont(new java.awt.Font("Dialog",
				java.awt.Font.BOLD, 14));
		this.sliderFitValue.setFont(new java.awt.Font("Dialog",
				java.awt.Font.BOLD, 14));
		this.sliderValue.setFont(new java.awt.Font("Dialog",
				java.awt.Font.BOLD, 14));

		this.sliderMinValue.setForeground(Color.GRAY);
		this.sliderMaxValue.setForeground(Color.GRAY);
		this.sliderFitValue.setForeground(Color.GRAY);
		this.sliderValue.setForeground(Color.DARK_GRAY);

		this.add(this.slider);
		this.add(this.sliderMinValue);
		this.add(this.sliderMaxValue);
		this.add(this.sliderFitValue);
		this.add(this.sliderValue);

		this.setBackground(Color.LIGHT_GRAY);
	}

	public void setHeight(int height) {

		this.setSize(115, height);

		int sliderHeight = height - 60;

		// this.title.setBounds(0, (int) (height * 0.5) - 25, 30, 50);

		this.slider.setBounds(35, 30, 30, sliderHeight);
		this.sliderMaxValue.setBounds(0, 10, 100, 20);
		this.sliderMinValue.setBounds(0, height - 30, 100, 20);

		int value = this.slider.getValue();
		int span = this.slider.getMaximum() - this.slider.getMinimum();
		int position = 33 + (int) ((float) (this.slider.getMaximum() - this.fitZoom)
				/ (float) span * (sliderHeight - 28));
		this.sliderFitValue.setBounds(0, position, 40, 20);

		if (value == this.fitZoom)
			this.sliderValue.setBounds(65, position, 60, 20);
		else {

			position = 33 + (int) ((float) (this.slider.getMaximum() - value)
					/ (float) span * (sliderHeight - 28));
			this.sliderValue.setBounds(65, position, 60, 20);
		}
	}

	private void update() {

		int value = this.slider.getValue();

		int span = this.slider.getMaximum() - this.slider.getMinimum();
		int position = 33 + (int) ((float) (this.slider.getMaximum() - value)
				/ (float) span * (this.slider.getBounds().height - 28));

		this.sliderValue.setText(value + "%");
		this.sliderValue.setBounds(65, position, 60, 20);
	}

	public void setValue(int value) {
		this.slider.setValue(value);
	}

	public void setFitValue(int value) {

		this.fitZoom = value;

		int span = this.slider.getMaximum() - this.slider.getMinimum();
		int position = 33 + (int) ((float) (this.slider.getMaximum() - value)
				/ (float) span * (this.slider.getBounds().height - 28));
		this.sliderFitValue.setBounds(0, position, 40, 20);
	}

	public void addSliderChangeListener(ChangeListener listener) {
		this.slider.addChangeListener(listener);
	}
}

class CnetInputOutput extends JPanel {
	
	private static final long serialVersionUID = 6000291731279974755L;
	private JScrollPane scrollPane;
	private JScrollPane scrollPaneText;
	private JTable table;
	private JTextArea textArea;
	
	public CnetInputOutput(HeuristicsNet net, String text){	
		table = InputOutputFunctions(net);
		
		scrollPane = new JScrollPane(table);
		textArea = new JTextArea(text);
		scrollPaneText = new JScrollPane(textArea);
//		Parent Bounds(10, 10, 520, 390);
		scrollPane.setBounds(0, 0, 520, 250);
		scrollPaneText.setBounds(0, 260, 520, 110);
		this.add(scrollPane);
		this.add(scrollPaneText);
		this.setLayout(null);
	}
	
	private JTable InputOutputFunctions(HeuristicsNet net) {
		Map<XEventClass, HNSubSet> activityMapping = net.getActivitiesMappingStructures().getReverseActivitiesMapping();
		Object[][] res = new Object[activityMapping.keySet().size()][3];
		
		for(XEventClass eventClass : activityMapping.keySet()) {
			int i = activityMapping.get(eventClass).get(0);
			String in = "";
			String out = "";
			HNSet outputset = net.getOutputSet(i);
			for(int j = 0; j < outputset.size(); j++) {
				out = out + "[ ";
				HNSubSet outss = outputset.get(j);
				for(int k = 0; k < outss.size(); k++) {
					out = out + outss.get(k) + " ";
				}
				out = out + "]";
			}
			
			HNSet inputset = net.getInputSet(i);
			for(int j = 0; j < inputset.size(); j++) {
				in = in + "[ ";
				HNSubSet inss = inputset.get(j);
				for(int k = 0; k < inss.size(); k++) {
					in = in + inss.get(k) + " ";
				}
				in = in + "]";
			}
			
			res[i][1] = i + " - " + eventClass.toString();
			res[i][0] = in;
			res[i][2] = out;
		}
		Object columnNames[] = {"Input", "Activity", "Output"};
		return new JTable(res, columnNames);
	}
	
	public void redrawTable(HeuristicsNet net) {
		table = InputOutputFunctions(net);
		this.validate();
		this.repaint();
	}
}

class VerticalTableHeaderCellRenderer extends
		javax.swing.table.DefaultTableCellRenderer {

	/**
	 * 
	 */
	private static final long serialVersionUID = -5792846837172592507L;

	public VerticalTableHeaderCellRenderer() {

		setOpaque(false);

		setHorizontalAlignment(LEFT);
		setHorizontalTextPosition(CENTER);
		setVerticalAlignment(CENTER);
		setVerticalTextPosition(TOP);
		setUI(new VerticalLabelUI(false));
	}

	protected javax.swing.Icon getIcon(JTable table, int column) {

		javax.swing.RowSorter.SortKey sortKey = getSortKey(table, column);
		if (sortKey != null && sortKey.getColumn() == column) {
			javax.swing.SortOrder sortOrder = sortKey.getSortOrder();
			switch (sortOrder) {
			case ASCENDING:
				return VerticalSortIcon.ASCENDING;
			case DESCENDING:
				return VerticalSortIcon.DESCENDING;
			}
		}
		return null;
	}

	private enum VerticalSortIcon implements javax.swing.Icon {

		ASCENDING, DESCENDING;
		private javax.swing.Icon icon = javax.swing.UIManager
				.getIcon("Table.ascendingSortIcon");

		/**
		 * Paints an icon suitable for the header of a sorted table column,
		 * rotated by 90° clockwise. This rotation is applied to compensate the
		 * rotation already applied to the passed in Graphics reference by the
		 * VerticalLabelUI.
		 * <P>
		 * The icon is retrieved from the UIManager to obtain an icon
		 * appropriate to the L&F.
		 * 
		 * @param c
		 *            the component to which the icon is to be rendered
		 * @param g
		 *            the graphics context
		 * @param x
		 *            the X coordinate of the icon's top-left corner
		 * @param y
		 *            the Y coordinate of the icon's top-left corner
		 */
		public void paintIcon(java.awt.Component c, Graphics g, int x, int y) {
			switch (this) {
			case ASCENDING:
				icon = javax.swing.UIManager.getIcon("Table.ascendingSortIcon");
				break;
			case DESCENDING:
				icon = javax.swing.UIManager
						.getIcon("Table.descendingSortIcon");
				break;
			}
			int maxSide = Math.max(getIconWidth(), getIconHeight());
			Graphics2D g2 = (Graphics2D) g.create(x, y, maxSide, maxSide);
			g2.rotate((Math.PI / 2));
			g2.translate(0, -maxSide);
			icon.paintIcon(c, g2, 0, 0);
			g2.dispose();
		}

		/**
		 * Returns the width of the rotated icon.
		 * 
		 * @return the <B>height</B> of the contained icon
		 */
		public int getIconWidth() {
			return icon.getIconHeight();
		}

		/**
		 * Returns the height of the rotated icon.
		 * 
		 * @return the <B>width</B> of the contained icon
		 */
		public int getIconHeight() {
			return icon.getIconWidth();
		}
	}

	protected javax.swing.RowSorter.SortKey getSortKey(JTable table, int column) {
		javax.swing.RowSorter<?> rowSorter = table.getRowSorter();
		if (rowSorter == null) {
			return null;
		}

		List<?> sortedColumns = rowSorter.getSortKeys();
		if (sortedColumns.size() > 0) {
			return (javax.swing.RowSorter.SortKey) sortedColumns.get(0);
		}
		return null;
	}

	@Override
	public java.awt.Component getTableCellRendererComponent(JTable table,
			Object value, boolean isSelected, boolean hasFocus, int row,
			int column) {
		super.getTableCellRendererComponent(table, value, isSelected, hasFocus,
				row, column);
		setIcon(getIcon(table, column));
		setBorder(null);
		return this;
	}
}

class DashedBorder extends javax.swing.border.LineBorder {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1357931293759243135L;

	public DashedBorder(Color color) {
		super(color);
	}

	public void paintBorder(java.awt.Component comp, Graphics g, int x1,
			int x2, int y1, int y2) {

		Stroke old = ((Graphics2D) g).getStroke();
		BasicStroke bs = new BasicStroke(5.0f, BasicStroke.CAP_BUTT,
				BasicStroke.JOIN_MITER, 10.0f, new float[] { 15.0f, 30.0f },
				2.0f);
		((Graphics2D) g).setStroke(bs);
		super.paintBorder(comp, g, x1, x2, y1, y2);
		((Graphics2D) g).setStroke(old);
	}
}