package com.processmininguc.feedbackrepair.plugins;

import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.heuristics.HeuristicsNet;

import com.processmininguc.feedbackrepair.models.FeedbackRepair;
import com.processmininguc.feedbackrepair.parameters.RepairHeuristicNetBasedOnFeedbackConfiguration;

@Plugin(
		name = "Repair Heuristic Net Based on Feedback", 
		parameterLabels = { "Heuristics Net", "Log", "Configuration" },  
		returnLabels = { "Feedback Repair" },
		returnTypes = { FeedbackRepair.class },
		categories = {PluginCategory.Enhancement},
		keywords = {"Enhancement", "Repair", "Causal Nets", "RPST", "Feedback"},
		help = "Allows the interactive repair of Heuristic Net based on user feedback",
		userAccessible = true
		)
public class RepairHeuristicNetBasedOnFeedbackPlugin {

	private FeedbackRepair repairPrivatePlugin(PluginContext context, HeuristicsNet net, XLog log, 
			RepairHeuristicNetBasedOnFeedbackConfiguration configuration){
				return new FeedbackRepair(net,log,configuration);
	}
	
	
	@PluginVariant(variantLabel = "Repair Heuristic Net Based on Feedback - configuration", requiredParameterLabels = { 0, 1, 2})
	public FeedbackRepair repairConfiguredPlugin(PluginContext context, HeuristicsNet net, XLog log, 
			RepairHeuristicNetBasedOnFeedbackConfiguration configuration){
	    return repairPrivatePlugin(context, net, log, configuration);
	}
	
	
	@UITopiaVariant(
			affiliation = "Universidad Catolica de Chile (UC)", 
			author = "Nicolas Pizarro", 
			email = "njpizarr"+ (char) 0x40 + "uc.cl",
			uiLabel = UITopiaVariant.USEPLUGIN,
			pack = "Feedback"
			)
	@PluginVariant(variantLabel = "Repair Heuristic Net Based on Feedback - default", requiredParameterLabels = { 0, 1})
	public FeedbackRepair repairDefaultPlugin(PluginContext context, HeuristicsNet net, XLog log){
		RepairHeuristicNetBasedOnFeedbackConfiguration configuration = 
				new RepairHeuristicNetBasedOnFeedbackConfiguration();
	    return repairPrivatePlugin(context, net, log, configuration);
	}	
}
